import json
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

from ...bliss_globals import current_session  # type: ignore
from ...bliss_globals import setup_globals  # type: ignore
from ...id31.streamline_scanner import Id31StreamlineScanner
from ...utils import directories
from ..processors.utils.pyfai_calib import DEFAULT_CALIB
from .utils.id31 import ensure_difflab6_id31_flats
from .utils.streamline import DemoStreamlineScannerMixIn


class DemoId31StreamlineScanner(DemoStreamlineScannerMixIn, Id31StreamlineScanner):
    _TEST_DATA_PATHS_WITHOUT_CALIB = (
        "/results/difflab6_integrate_q/integrated/q",
        "/results/difflab6_integrate_2th/integrated/2th",
        "/results/difflab6_integrate_q_no_sigmaclip/integrated/q",
    )
    _TEST_DATA_PATHS_WITH_CALIB = (
        "/results/difflab6_integrate_q/integrated/q",
        "/results/difflab6_integrate_2th/integrated/2th",
    )
    _TEST_ASCII_SUFFIXES_WITHOUT_CALIB = ("_2th.xye", "_q.xye")
    _TEST_ASCII_SUFFIXES_WITH_CALIB = ("_2th.xye", "_q.xye")

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ):
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("optimize_exposure_per", "sample")
        defaults.setdefault("default_attenuator", 4)
        defaults.setdefault("energy_name", "energy")
        defaults.setdefault("auto_stop_threshold", 0.01)
        defaults.setdefault("detector_saturation", 1e6)
        defaults.setdefault("auto_stop_scan_npoints", 100)

        id31_patching()

        super().__init__(config=config, defaults=defaults)

        if self._HAS_BLISS:
            self.newflat, self.oldflat = ensure_difflab6_id31_flats()

    def init_auto_stop_workflow(self) -> None:
        """Initialise the demo auto-stop workflow.

        Creates the PyFAI config file if needed, writes the
        ``PyFaiConfig → IntegrateAndSendMsg`` Ewoks workflow JSON into the
        proposal's workflows directory, and sets ``auto_stop_workflow_path``.
        """
        from .stop_scan_xrpd_id31 import _ensure_demo_pyfai_config

        self.pyfai_config = _ensure_demo_pyfai_config()

        integrate_identifier = (
            "blissoda.demo.tasks.stop_scan_xrpd_integrate.IntegrateAndSendMsg"
        )
        write_sum_identifier = (
            "blissoda.demo.tasks.stop_scan_xrpd_integrate.WriteSumFrame"
        )
        workflow = {
            "graph": {"schema_version": "1.1", "id": "streamline_auto_stop"},
            "nodes": [
                {
                    "id": "config",
                    "task_type": "class",
                    "task_identifier": "ewoksxrpd.tasks.pyfaiconfig.PyFaiConfig",
                },
                {
                    "id": "integrate",
                    "task_type": "class",
                    "task_identifier": integrate_identifier,
                },
                {
                    "id": "write_sum",
                    "task_type": "class",
                    "task_identifier": write_sum_identifier,
                },
            ],
            "links": [
                {
                    "source": "config",
                    "target": "integrate",
                    "data_mapping": [
                        {
                            "source_output": "detector_config",
                            "target_input": "detector_config",
                        },
                        {"source_output": "energy", "target_input": "energy"},
                        {
                            "source_output": "integration_options",
                            "target_input": "integration_options",
                        },
                        {"source_output": "detector", "target_input": "detector"},
                        {"source_output": "mask", "target_input": "mask"},
                        {"source_output": "geometry", "target_input": "geometry"},
                    ],
                },
                {
                    "source": "integrate",
                    "target": "write_sum",
                    "data_mapping": [],
                },
            ],
        }
        workflows_dir = directories.get_workflows_dir(
            current_session.scan_saving.filename
        )
        Path(workflows_dir).mkdir(parents=True, exist_ok=True)
        workflow_path = Path(workflows_dir) / "streamline_auto_stop.json"
        workflow_path.write_text(json.dumps(workflow, indent=2))
        self.auto_stop_workflow_path = str(workflow_path)
        print(f"Auto-stop workflow initialized: {workflow_path}")


def id31_patching(energy: Optional[float] = None):
    if "id31" in sys.modules:
        return

    from ...tests import mock_id31

    # Patch modules to allow "from id31 import attenuator"
    sys.modules["id31"] = mock_id31
    sys.modules["id31.attenuator"] = mock_id31.attenuator

    from ...tests.mock_id31.setup_globals import atten
    from ...tests.mock_id31.setup_globals import ehss

    setup_globals.shopen = _mock_shopen
    if energy is None:
        setup_globals.energy.move(DEFAULT_CALIB["difflab6"]["energy"])
    else:
        setup_globals.energy.move(energy)
    setup_globals.atten = atten
    setup_globals.att = _att
    setup_globals.ehss = ehss

    from ...id31 import optimize_exposure

    optimize_exposure.id31_attenuator = mock_id31.attenuator

    # Keep it patched for manual testing


def _mock_shopen(**kwargs):
    arguments = ", ".join(f"{k}={v}" for k, v in kwargs.items())
    print(f"shopen({arguments})")


def _att(value):
    setup_globals.atten.bits = value



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from pydantic import BaseModel

from . import defaults
from . import detectors as detectors_module


class XrfMapPaths(BaseModel):
    filename: str
    output_root_uri: str
    convert_destination: str
    workflow_path: str
    config_filenames: List[str]


class MosaicXrfMapPaths(BaseModel):
    filenames: List[str]
    output_root_uri: str
    concat_bliss_scan_uri: str
    convert_destination: str
    workflow_path: str
    config_filenames: List[str]


class FluoXasPaths(BaseModel):
    filenames: List[str]
    output_root_uri: str
    convert_destination: str
    workflow_path: str
    config_filenames: List[str]


class CommonParameters(BaseModel):
    instrument_name: Optional[str]
    detector_names: List[str]
    detector_normalization_template: str
    counter_name: Optional[str]
    counter_normalization_template: Optional[str]
    config_filenames: List[str]
    energy_name: Optional[str]
    energy_multiplier: Optional[float]
    fast_fitting: bool
    quantification: bool
    diagnostics: bool
    no_fit: bool
    fit_single_detector: Optional[bool]
    sum_spectra: Optional[bool]
    norm_identifier: str
    fit_identifier: str
    virtual_axes: Optional[Dict[str, str]]
    ignore_axes: Optional[List[str]]
    axis_units: Optional[Dict[str, str]]

    @property
    def regrid_positioners(self) -> List[str]:
        return sorted(self.virtual_axes)[::-1]


class XrfMapParameters(CommonParameters):
    filename: str
    scan_number: int
    output_root_uri: str
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]]


class MosaicXrfMapParameters(CommonParameters):
    filenames: List[str]
    scan_ranges: Sequence[Tuple[int, int]]
    exclude_scans: Sequence[Sequence[int]]
    output_root_uri: str
    bliss_scan_uri: str
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]]


class FluoXasParameters(CommonParameters):
    filenames: List[str]
    scan_ranges: Sequence[Tuple[int, int]]
    exclude_scans: Sequence[Sequence[int]]
    output_root_uri: str
    resolution: Optional[Dict[str, Tuple[Union[int, float], str]]]
    align_counter: Optional[str]
    align_crop: Optional[bool]
    fast_align_counter_selection: Optional[bool]
    interpolation_order: int = 1
    transformation_type: str = "translation"
    block_size: int = 10
    mapper: str = "Optimization-SimpleITK"
    mapper_options: dict = {
        "metric": "ANTSNeighborhoodCorrelation",
        "optimizer": "Powell",
    }

    @property
    def skip_pre_evaluation(self) -> bool:
        return not self.align_counter and not self.fast_align_counter_selection

    @property
    def skip_post_evaluation(self) -> bool:
        return not self.align_counter and self.fast_align_counter_selection


def common_parameters_model(
    detectors: Optional[Sequence[Union[int, str]]],
    detector_numbers: Optional[Sequence[int]],
    detector_names: Optional[Sequence[str]],
    config_filenames: Sequence[str],
    energy_name: Optional[str],
    energy_multiplier: Optional[float],
    counter_name: Optional[str],
    instrument_name: Optional[str],
    fast_fitting: bool,
    quantification: bool,
    diagnostics: bool,
    livetime_ref_value: Union[str, int, float, None],
    counter_ref_value: Union[str, int, float, None],
    stack: bool,
    virtual_axes: Optional[Dict[str, str]],
    ignore_axes: Optional[List[str]],
    axis_units: Optional[Dict[str, str]],
) -> Tuple[Sequence[dict], Optional[bool]]:
    detector_names = detectors_module.get_detector_names(
        detectors=detectors,
        detector_numbers=detector_numbers,
        detector_names=detector_names,
        instrument_name=instrument_name,
    )

    # Default strings
    energy_name = (
        energy_name
        if energy_name is not None
        else defaults.ENERGY_COUNTER.get(instrument_name)
    )
    if energy_name == "":
        energy_name = None

    counter_name = (
        counter_name
        if counter_name is not None
        else defaults.I0_COUNTER.get(instrument_name)
    )
    if counter_name == "":
        counter_name = None

    # Default dicts
    axis_units = _merge_dict(axis_units, defaults.AXES_UNITS.get(instrument_name))

    ignore_axes = (
        ignore_axes
        if ignore_axes is not None
        else defaults.IGNORE_AXES.get(instrument_name)
    )

    virtual_axes = (
        virtual_axes
        if virtual_axes is not None
        else defaults.VIRTUAL_AXES.get(instrument_name)
    )

    # Normalization templates: None gives the default, "" gives None
    detector_normalization_template = _build_normalization_template(
        livetime_ref_value,
        defaults.DEFAULT_LIVETIME_REF_VALUE.get(instrument_name),
        "live_time",
    )
    counter_normalization_template = _build_normalization_template(
        counter_ref_value,
        defaults.DEFAULT_COUNTER_REF_VALUE.get(instrument_name),
        "data",
    )

    if counter_name is None:
        counter_normalization_template = None
    elif counter_normalization_template is None:
        counter_name = None

    # Number of detectors vs. number of configuration files
    if len(detector_names) == 0 or len(config_filenames) == 0:
        no_fit = True
        sum_spectra = None
        fit_single_detector = None
    elif len(detector_names) == 1:
        # Only one detector
        if len(config_filenames) != 1:
            raise ValueError("Only one pymca configuration is needed for one detector")
        no_fit = False
        sum_spectra = None
        fit_single_detector = True
    elif len(config_filenames) == 1:
        # More than one detector and fit the sum
        no_fit = False
        sum_spectra = True
        fit_single_detector = True
    else:
        # More than one detector and fit each detector separately
        if len(detector_names) != len(config_filenames):
            raise ValueError(
                f"{len(detector_names)} pymca configurations are needed for {len(detector_names)} detectors"
            )
        no_fit = False
        sum_spectra = False
        fit_single_detector = False

    if stack:
        norm_identifier = "NormalizeXrfResultsStack"
        if fit_single_detector:
            fit_identifier = "FitStackSingleDetector"
        else:
            fit_identifier = "FitStackMultiDetector"
    else:
        norm_identifier = "NormalizeXrfResults"
        if fit_single_detector:
            fit_identifier = "FitSingleScanSingleDetector"
        else:
            fit_identifier = "FitSingleScanMultiDetector"

    return CommonParameters(
        instrument_name=instrument_name,
        detector_names=detector_names,
        detector_normalization_template=detector_normalization_template,
        counter_name=counter_name,
        counter_normalization_template=counter_normalization_template,
        energy_name=energy_name,
        energy_multiplier=energy_multiplier,
        fast_fitting=fast_fitting,
        quantification=quantification,
        diagnostics=diagnostics,
        no_fit=no_fit,
        fit_single_detector=fit_single_detector,
        config_filenames=config_filenames,
        sum_spectra=sum_spectra,
        norm_identifier=norm_identifier,
        fit_identifier=fit_identifier,
        virtual_axes=virtual_axes,
        ignore_axes=ignore_axes,
        axis_units=axis_units,
    )


def _build_normalization_template(
    value: Union[str, int, float, None],
    default_value: Union[str, int, float, None],
    dataset: str,
) -> Optional[str]:
    if value is None:
        value = default_value
    if value is not None and value != "":
        return f"{value}/<instrument/{{}}/{dataset}>"
    return None


def _merge_dict(
    adict: Optional[Dict[str, Any]], default: Optional[Dict[str, Any]]
) -> Optional[Dict[str, Any]]:
    if not adict and not default:
        return None

    result = {}

    if default:
        result.update(default)
    if adict:
        result.update(adict)

    return result



from __future__ import annotations

import sys


def ensure_numpy_pickle_compat() -> None:
    """Allow NumPy 1 environments to unpickle arrays serialized by NumPy 2.

    NumPy 2 moved array implementation modules from ``numpy.core`` to
    ``numpy._core``. When a worker running NumPy 2 returns pickled arrays to a
    BLISS process still running NumPy 1, unpickling can fail because the older
    environment does not provide the ``numpy._core`` module path.
    """
    try:
        import numpy.core as numpy_core
    except Exception:
        return

    sys.modules.setdefault("numpy._core", numpy_core)

    try:
        import numpy.core.multiarray as numpy_core_multiarray
    except Exception:
        return

    sys.modules.setdefault("numpy._core.multiarray", numpy_core_multiarray)



import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ..persistent.parameters import ParameterInfo
from ..resources import resource_filename
from ..utils.pyfai import read_config
from ..xrpd.processor import BlissScanType
from ..xrpd.processor import XrpdProcessor
from ..xrpd.processor import _get_scan_memory_url

logger = logging.getLogger(__name__)

SLURM_JOB_PARAMETERS_INTEGRATE = {
    "name": "ID13_pyFAI-integrate",
    "partition": "gpu",
    "time": "01:00:00",
    "tasks_per_node": 1,
    "cpus_per_task": 1,
    "memory_per_cpu": "50G",
    "tres_per_job": "gres/gpu:1",
    "constraints": "l40s",
}

SLURM_JOB_PARAMETERS_BACKGROUND = {
    "name": "ID13_spi-background",
    "partition": "nice",
    "time": "02:00:00",
    "tasks_per_node": 1,
    "cpus_per_task": 32,
    "memory_per_cpu": "2G",
    "tres_per_job": None,
    "constraints": None,
}

SLURM_JOB_PARAMETERS_CNMF = {
    "name": "ID13_spi-CNMF",
    "partition": "gpu",
    "time": "01:00:00",
    "tasks_per_node": 1,
    "cpus_per_task": 1,
    "memory_per_cpu": "50G",
    "tres_per_job": "gres/gpu:1",
    "constraints": "l40s",
}

QUEUES_GPU_ID13 = [
    "gpu3",
    "gpu2",
]
QUEUE_SLURM_ID13 = "slurm"
WORKER_MODULE = "scattering"
PRE_SCRIPT = "module load {WORKER_MODULE}; python3 -m ewoksid13.scripts.utils.slurm_python_pre_script"
PYTHON_CMD = "python3"
POST_SCRIPT = "python3 -m ewoksid13.scripts.utils.slurm_python_post_script"


class Id13XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("workflow_juno", category="workflows"),
        ParameterInfo("current_workflow", category="workflows"),
        ParameterInfo("normalization_counter", category="Processing"),
        ParameterInfo("do_diffmap", category="Processing"),
        ParameterInfo("do_average", category="Processing"),
        ParameterInfo("do_stackedf", category="Processing"),
        ParameterInfo("average_reference", category="Processing"),
        ParameterInfo("save_external_files", category="Processing"),
        ParameterInfo("radial_limits", category="Processing"),
        ParameterInfo("directory_cif_phases", category="Processing"),
        ParameterInfo("do_background_removal", category="Processing"),
        ParameterInfo("do_cnmf", category="Processing"),
        ParameterInfo("nb_components", category="Processing"),
    ],
):
    DEFAULT_LIMA_URL_TEMPLATE: Optional[str] = (
        "{dirname}/scan{scan_number_as_str}/{images_prefix}{{file_index}}.h5::/entry_0000/measurement/data"
    )
    DEFAULT_WORKFLOW_JUNO: Optional[str] = resource_filename("id13", "juno.json")

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
        **deprecated_defaults: Dict[str, Any],
    ) -> None:
        defaults = self._merge_defaults(deprecated_defaults, defaults)

        defaults.setdefault("current_workflow", "")
        defaults.setdefault("save_scans_separately", True)
        defaults.setdefault("do_diffmap", False)
        defaults.setdefault("normalization_counter", None)
        defaults.setdefault("do_average", False)
        defaults.setdefault("do_stackedf", False)
        defaults.setdefault("average_reference", "hydrocerussite")
        defaults.setdefault("save_external_files", False)
        defaults.setdefault("workflow_juno", self.DEFAULT_WORKFLOW_JUNO)
        defaults.setdefault("do_background_removal", False)
        defaults.setdefault("radial_limits", None)
        defaults.setdefault("do_cnmf", False)
        defaults.setdefault("directory_cif_phases", None)
        defaults.setdefault("nb_components", None)

        self._do_phase_inference = False  # Not available for online
        self._inference_weights_filename = None
        self._submit_slurm_integration = False
        self._slurm_module = WORKER_MODULE
        self._slurm_pre_script = PRE_SCRIPT
        self._slurm_python_cmd = PYTHON_CMD
        self._slurm_post_script = POST_SCRIPT
        self._slurm_parameters_integration = SLURM_JOB_PARAMETERS_INTEGRATE
        self._slurm_parameters_background = SLURM_JOB_PARAMETERS_BACKGROUND
        self._slurm_parameters_cnmf = SLURM_JOB_PARAMETERS_CNMF
        self._slurm_queue = QUEUE_SLURM_ID13
        self._queues = QUEUES_GPU_ID13
        super().__init__(config=config, defaults=defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[dict]:
        if self.pyfai_config:
            integration_options = read_config(filename=self.pyfai_config)
        else:
            integration_options = dict()
        if self.integration_options:
            integration_options.update(self.integration_options.to_dict())
        return integration_options

    def _get_lima_url_template_args(
        self, scan: BlissScanType, lima_name: str
    ) -> Optional[Dict[str, str]]:
        if self.lima_url_template_args:
            lima_url_template_args = dict(self.lima_url_template_args)
        else:
            lima_url_template_args = dict()
        eval_dict = {"img_acq_device": lima_name, "scan_number": scan.scan_number}
        images_prefix = scan.scan_saving.images_prefix.format(**eval_dict)
        lima_url_template_args["images_prefix"] = images_prefix
        lima_url_template_args["scan_number_as_str"] = scan.scan_number
        return lima_url_template_args

    def get_external_output_filename(self, scan: BlissScanType, lima_name: str):
        external_output_filename = self.external_output_filename(
            scan=scan, lima_name=lima_name
        )
        if not external_output_filename:
            scan_nb = scan.scan_info.get("scan_nb")
            master_output_filename = self.master_output_filename(scan)
            external_output_filename = master_output_filename.replace(
                ".h5", f"_{scan_nb:04d}.h5"
            )
        return external_output_filename

    def get_inputs(self, scan: BlissScanType, lima_name: str) -> List[dict]:
        """Additional inputs for the Id13 XrpdProcessor."""
        inputs = super().get_inputs(scan=scan, lima_name=lima_name)
        inputs += self.get_diff_extra_inputs(scan=scan, lima_name=lima_name)
        inputs += self.get_neuralnetwork_inputs(scan=scan, lima_name=lima_name)
        return inputs

    def get_diff_extra_inputs(
        self,
        scan: BlissScanType,
        lima_name: str,
    ) -> List[dict]:
        """Get additional inputs for the diffmap, average, and stacked EDF tasks."""
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        master_output_filename = self.master_output_filename(scan)
        external_output_filename = self.get_external_output_filename(
            scan=scan, lima_name=lima_name
        )
        external_output_filename_edf = external_output_filename.replace(".h5", ".edf")

        if self.save_external_files:
            external_output_filename = external_output_filename
        else:
            external_output_filename = master_output_filename

        params = {
            "CreateDiffMapFile": {
                "master_filename": filename,
                "scan_nb": scan_nb,
                "output_filename": master_output_filename,
                "external_output_filename": external_output_filename,
                "integration_options": self.get_integration_options(
                    scan=scan, lima_name=lima_name
                ),
                "lima_name": lima_name,
                "do_diffmap": self.do_diffmap,
                "scan_memory_url": _get_scan_memory_url(scan),
                "normalization_counter": self.normalization_counter,
                "processing_info": self.get_saving_info(scan),
            },
            "AverageIntegration": {
                "master_filename": filename,
                "scan_nb": scan_nb,
                "lima_name": lima_name,
                "output_filename": master_output_filename,
                "external_output_filename": external_output_filename,
                "do_average": self.do_average,
                "reference": self.average_reference,
                "scan_memory_url": _get_scan_memory_url(scan),
                "normalization_counter": self.normalization_counter,
                "processing_info": self.get_saving_info(scan),
            },
            "StackToEdf": {
                "do_stackedf": self.do_stackedf,
                "external_output_filename": external_output_filename_edf,
                "processing_info": self.get_saving_info(scan),
            },
        }

        return [
            {"name": name, "value": value, "task_identifier": task_id}
            for task_id, task_params in params.items()
            for name, value in task_params.items()
        ]

    def get_neuralnetwork_inputs(
        self, scan: BlissScanType, lima_name: str
    ) -> List[dict]:
        """Get additional inputs for the neural network (spi) tasks: background removal, constrained NMF and phase inference."""
        external_output_filename = self.get_external_output_filename(
            scan=scan, lima_name=lima_name
        )

        wait_for_bg = self.do_cnmf
        do_background_removal = self.do_background_removal or self.do_cnmf

        common_params = {
            "wavelength": self.get_wavelength(angstroms=True),
            "radial_limits": self.radial_limits,
            "worker_module": self._slurm_module,
            "processing_info": self.get_saving_info(scan),
        }

        if not do_background_removal:
            params = {
                "BackgroundRemoval": {
                    "do_background_removal": False,
                    **common_params,
                },
                "ConstrainedNMF": {
                    "do_matrix_factorization": False,
                    **common_params,
                },
                "PhaseInference": {
                    "do_phase_inference": False,
                    "references_directory": self.directory_cif_phases,
                    **common_params,
                },
            }
        else:
            params = {
                "BackgroundRemoval": {
                    "do_background_removal": True,
                    "force_training": False,  # Not good for online processing
                    "use_neuralnetwork": False,  # Not good for online processing
                    "submit_to_slurm": True,  # Background needs multiprocessing (--pool=process, which appears to jeopardize the integration sometimes)
                    "destination_file": external_output_filename.replace(
                        ".h5", "_background.json"
                    ),
                    "slurm_job_parameters": self._slurm_parameters_background,
                    "wait_to_finish": wait_for_bg,
                    **common_params,
                },
                "ConstrainedNMF": {
                    "references_directory": self.directory_cif_phases,
                    "nb_components": self.nb_components,
                    "do_matrix_factorization": self.do_cnmf,
                    "submit_to_slurm": True,  # This task is too heavy to leave it running on gpu3 during the experiment
                    "destination_file": external_output_filename.replace(
                        ".h5", "_cnmf.json"
                    ),
                    "slurm_job_parameters": self._slurm_parameters_cnmf,
                    "wait_to_finish": False,
                    **common_params,
                },
                "PhaseInference": {
                    "do_phase_inference": self._do_phase_inference,
                    "references_directory": self.directory_cif_phases,
                    "inference_weights_filename": self._inference_weights_filename,
                    **common_params,
                },
            }

        return [
            {"name": name, "value": value, "task_identifier": task_id}
            for task_id, task_params in params.items()
            for name, value in task_params.items()
        ]

    def get_wavelength(self, angstroms: bool = True) -> float:
        """Load the wavelength from the PyFai configuration file."""
        with open(self.pyfai_config) as f:
            config = json.load(f)
        if "wavelength" in config:
            return float(config["wavelength"]) * (1e10 if angstroms else 1)
        if "poni" in config:
            return float(config["poni"]["wavelength"]) * (1e10 if angstroms else 1)

    def _get_workflow(self, scan: BlissScanType) -> Optional[str]:
        """Get the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            self.current_workflow = self.workflow_juno
        else:
            self.current_workflow = self.workflow_without_saving
        return self.current_workflow

    def _set_workflow(self, scan: BlissScanType, filename: str) -> None:
        """Set the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            self.workflow_with_saving = filename
        else:
            self.workflow_without_saving = filename
        self.current_workflow = filename

    def get_submit_arguments(self, scan: BlissScanType, lima_name) -> dict:
        """Additional submit arguments in case of slurm queue, also switch the queue if needed."""
        submit_arguments = super().get_submit_arguments(scan, lima_name)
        if self.queue == self._slurm_queue:
            slurm_parameters = {
                **self._slurm_parameters_integration,
                "environment": {
                    "HOME": "/tmp_14_days/dau_id13_opid13",
                },
            }
            return {
                **submit_arguments,
                "slurm_arguments": {
                    "parameters": slurm_parameters,
                    "pre_script": self._slurm_pre_script.format(
                        WORKER_MODULE=self._slurm_module
                    ),
                    "python_cmd": self._slurm_python_cmd,
                    "post_script": self._slurm_post_script,
                },
            }
        return submit_arguments

    def get_saving_info(self, scan: BlissScanType) -> dict:
        return {
            "base_path": scan.scan_saving.base_path,
            "template": scan.scan_saving.template,
            "proposal_dirname": scan.scan_saving.proposal_dirname,
            "beamline": scan.scan_saving.beamline,
            "proposal_session_name": scan.scan_saving.proposal_session_name,
            "collection_name": scan.scan_saving.collection_name,
            "dataset_name": scan.scan_saving.dataset_name,
        }



from __future__ import annotations

import json
import logging
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import gevent
from esrf_pathlib import ESRFPath
from ewoksjob.client import get_future
from ewoksjob.client import submit

from ..persistent.parameters import ParameterInfo
from ..processor import BaseProcessor
from ..processor import BlissScanType
from ..resources import resource_filename
from .xrr_plotter import Bm32XrrPlotter

logger = logging.getLogger(__name__)


class Bm32XrrProcessor(
    BaseProcessor,
    parameters=[
        ParameterInfo("queue", category="workflows"),
        ParameterInfo(
            "workflow", category="workflows", doc="Main XRR processing workflow (.json)"
        ),
        ParameterInfo("data_portal_upload", category="data portal"),
        ParameterInfo("plot_name", category="plotting"),
        ParameterInfo(
            "tth_path",
            category="XRR (dataset paths)",
            doc="e.g. '{scan}.1/measurement/tth'",
        ),
        ParameterInfo(
            "det_path",
            category="XRR (dataset paths)",
            doc="e.g. '{scan}.1/measurement/det'",
        ),
        ParameterInfo(
            "mon_path",
            category="XRR (dataset paths)",
            doc="e.g. '{scan}.1/measurement/mon4'",
        ),
        ParameterInfo(
            "energy", category="Corrections/Geometry", doc="X-ray energy (eV)"
        ),
        ParameterInfo(
            "itime", category="Corrections/Geometry", doc="Integration time (seconds)"
        ),
        ParameterInfo(
            "dark_current",
            category="Corrections/Geometry",
            doc="Dark current (scalar or 1D)",
        ),
        ParameterInfo(
            "direct_beam_amp",
            category="Corrections/Geometry",
            doc="Direct beam amplitude (fit prior)",
        ),
        ParameterInfo(
            "beam_size", category="Corrections/Geometry", doc="Beam size on sample (mm)"
        ),
        ParameterInfo(
            "sample_length",
            category="Corrections/Geometry",
            doc="Illuminated sample length (mm)",
        ),
        ParameterInfo(
            "tthmin",
            category="Corrections/Geometry",
            doc="Lower angle cutoff for processing (deg)",
        ),
        ParameterInfo(
            "qvaluesRangesToRemove",
            category="Clean Signal",
            doc="List of q-ranges to remove from the signal (e.g. [ [0.1,0.15], [0.25,0.3] ])",
        ),
        ParameterInfo(
            "qmax",
            category="Clean Signal",
            doc="Maximum q-value to consider (1/Angstrom)",
        ),
        ParameterInfo(
            "peak_prominence",
            category="Inverse FT",
            doc="Peak prominence for find_peaks (Inverse FT)",
        ),
        ParameterInfo(
            "baseline_type", category="Inverse FT", doc="Baseline type (Inverse FT)"
        ),
        ParameterInfo(
            "peak_distance",
            category="Inverse FT",
            doc="Minimum peak distance (Inverse FT)",
        ),
        ParameterInfo(
            "zrange", category="Inverse FT", doc="Z range to consider (Inverse FT)"
        ),
    ],
):
    """Trigger the BM32 XRR workflow from BLISS scan objects."""

    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}

        defaults.setdefault("workflow", "xrr_processor.json")
        defaults.setdefault("data_portal_upload", False)
        defaults.setdefault("plot_name", "XRR analysis")

        defaults.setdefault("tth_path", "{scan}.1/measurement/tth")
        defaults.setdefault("det_path", "{scan}.1/measurement/det")
        defaults.setdefault("mon_path", "{scan}.1/measurement/mon4")

        defaults.setdefault("itime", None)
        defaults.setdefault("dark_current", 0.0)
        defaults.setdefault("direct_beam_amp", None)
        defaults.setdefault("beam_size", None)
        defaults.setdefault("sample_length", None)
        defaults.setdefault("tthmin", None)
        defaults.setdefault("energy", None)
        defaults.setdefault("qvaluesRangesToRemove", [])
        defaults.setdefault("qmax", 0.8)
        defaults.setdefault("peak_prominence", 1e-10)
        defaults.setdefault("baseline_type", "pchip")
        defaults.setdefault("peak_distance", None)
        defaults.setdefault("zrange", 100)

        super().__init__(config=config, defaults=defaults)
        self._tasks: List[Any] = []
        self.plotter = Bm32XrrPlotter()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        categories.setdefault("status", {})["Workflow tasks"] = self.purge_tasks()
        if self.plotter is not None:
            categories.setdefault("status", {})[
                "Plotting tasks"
            ] = self.plotter.purge_tasks()
        return categories

    def purge_tasks(self) -> int:
        self._tasks = [task for task in self._tasks if task]
        return len(self._tasks)

    def _spawn(self, *args, **kwargs) -> None:
        self._tasks.append(gevent.spawn(*args, **kwargs))
        self.purge_tasks()

    @staticmethod
    def _consume_future_result(future, scan_nb: int) -> None:
        try:
            future.result()
        except Exception:
            logger.exception("BM32 XRR workflow failed for scan %s", scan_nb)

    @staticmethod
    def _scan_number_from_scan(scan: BlissScanType) -> int:
        return int(scan.scan_info["scan_nb"])

    @staticmethod
    def _filename_from_scan(scan: BlissScanType) -> str:
        return scan.scan_info["filename"]

    @staticmethod
    def _get_output_dir(master_filename: str) -> str:
        return str(ESRFPath(master_filename).processed_dataset_path)

    def _set_output_basename(
        self, scan: BlissScanType, bg_scan_nb: Optional[int] = None
    ) -> str:
        scan_saving = scan.scan_saving
        sample = scan_saving.sample_name or "sample"
        dataset = scan_saving.dataset_name or "dataset"
        scan_nb = self._scan_number_from_scan(scan)
        return f"{sample}_{dataset}_{scan_nb}_{bg_scan_nb}"

    def _ensure_ready(self, scan: BlissScanType) -> None:
        self._scan_number_from_scan(scan)
        master_filename = self._filename_from_scan(scan)
        if not master_filename:
            raise ValueError("master filename is required")

    def workflow_destination(
        self, scan: BlissScanType, background: Optional[BlissScanType] = None
    ) -> str:
        master_filename = self._filename_from_scan(scan)
        output_dir = self._get_output_dir(master_filename)
        bg_scan_nb = (
            self._scan_number_from_scan(background) if background is not None else None
        )
        output_basename = self._set_output_basename(scan, bg_scan_nb)
        return os.path.join(output_dir, f"{output_basename}.json")

    def export_destination(
        self, scan: BlissScanType, background: Optional[BlissScanType] = None
    ) -> str:
        master_filename = self._filename_from_scan(scan)
        output_dir = self._get_output_dir(master_filename)
        bg_scan_nb = (
            self._scan_number_from_scan(background) if background is not None else None
        )
        output_basename = self._set_output_basename(scan, bg_scan_nb)
        return os.path.join(output_dir, f"{output_basename}.hdf5")

    def get_inputs(
        self, scan: BlissScanType, background: Optional[BlissScanType] = None
    ) -> List[Dict[str, Any]]:
        scan_nb = self._scan_number_from_scan(scan)
        master_filename = self._filename_from_scan(scan)
        bg_master_filename = (
            self._filename_from_scan(background) if background is not None else ""
        )
        bg_scan_nb = (
            self._scan_number_from_scan(background) if background is not None else None
        )
        output_basename = self._set_output_basename(scan, bg_scan_nb)

        inputs: List[Dict[str, Any]] = [
            {
                "task_identifier": "ewoksbm32.tasks.xrr.load_raw_data.LoadRawData",
                "name": "filename",
                "value": master_filename,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.load_raw_data.LoadRawData",
                "name": "tth_path",
                "value": self.tth_path.format(scan=scan_nb),
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.load_raw_data.LoadRawData",
                "name": "det_path",
                "value": self.det_path.format(scan=scan_nb),
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.load_raw_data.LoadRawData",
                "name": "mon_path",
                "value": self.mon_path.format(scan=scan_nb),
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.load_raw_data.LoadRawData",
                "name": "bg_nx_data_url",
                "value": bg_master_filename or master_filename,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.load_raw_data.LoadRawData",
                "name": "itime",
                "value": self.itime,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "energy",
                "value": self.energy,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "itime",
                "value": self.itime,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "dark_current",
                "value": self.dark_current,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "direct_beam_amp",
                "value": self.direct_beam_amp,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "beam_size",
                "value": self.beam_size,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "sample_length",
                "value": self.sample_length,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.preprocess.Preprocess",
                "name": "tthmin",
                "value": self.tthmin,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.clean_signal.CleanSignal",
                "name": "qvaluesRangesToRemove",
                "value": self.qvaluesRangesToRemove,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.clean_signal.CleanSignal",
                "name": "qmax",
                "value": self.qmax,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.inverse_ft.InverseFT",
                "name": "peak_prominence",
                "value": self.peak_prominence,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.inverse_ft.InverseFT",
                "name": "baseline_type",
                "value": self.baseline_type,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.inverse_ft.InverseFT",
                "name": "peak_distance",
                "value": self.peak_distance,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.inverse_ft.InverseFT",
                "name": "zrange",
                "value": self.zrange,
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.export_data.ExportHDF5",
                "name": "filename",
                "value": self.export_destination(scan, background),
            },
            {
                "task_identifier": "ewoksbm32.tasks.xrr.export_data.ExportHDF5",
                "name": "sample_name",
                "value": output_basename,
            },
        ]

        return [item for item in inputs if item.get("value") is not None]

    def _get_workflow(self) -> Dict[str, Any]:
        with open(resource_filename("bm32", self.workflow), "r") as wf:
            return json.load(wf)

    def get_submit_arguments(
        self, scan: BlissScanType, background: Optional[BlissScanType] = None
    ) -> Dict[str, Any]:
        return {
            "inputs": self.get_inputs(scan, background),
            "outputs": [{"all": True}],
        }

    def _get_workflow_upload_parameters(self, scan: BlissScanType) -> Dict[str, Any]:
        master_filename = self._filename_from_scan(scan)
        output_dir = self._get_output_dir(master_filename)
        dataset_metadata = scan.scan_saving.dataset or {}
        sample_name = scan.scan_saving.sample_name
        esrf_path = ESRFPath(master_filename)
        raw_directory = esrf_path.raw_dataset_path
        beamline = raw_directory.beamline
        proposal = raw_directory.proposal
        raw_directory = str(raw_directory)
        metadata = {"Sample_name": dataset_metadata.get("Sample_name", sample_name)}
        return {
            "beamline": beamline,
            "proposal": proposal,
            "dataset": "XRR_processing",
            "path": output_dir,
            "raw": [raw_directory],
            "metadata": metadata,
        }

    def _trigger_plotting(
        self,
        future,
        scan: BlissScanType,
        background: Optional[BlissScanType] = None,
    ) -> None:
        if self.plotter is None:
            return

        scan_label = f"scan {self._scan_number_from_scan(scan)}"
        if background is not None:
            scan_label += f" (bg {self._scan_number_from_scan(background)})"

        scan_saving = scan.scan_saving
        sample = scan_saving.sample_name or "sample"
        dataset = scan_saving.dataset_name or "dataset"
        display_name = f"{sample} | {dataset} | {scan_label}"
        self.plotter.handle_workflow_result(
            future=future,
            sample_name=display_name,
            plot_name=self.plot_name,
        )

    def run_processing(
        self,
        scan: BlissScanType,
        background: Optional[BlissScanType] = None,
    ) -> None:
        self._ensure_ready(scan)
        workflow = self._get_workflow()
        kwargs = self.get_submit_arguments(scan, background)
        kwargs["convert_destination"] = self.workflow_destination(scan, background)
        if self.data_portal_upload:
            kwargs["upload_parameters"] = self._get_workflow_upload_parameters(scan)

        future = submit(args=(workflow,), kwargs=kwargs, queue=self.queue)
        future = get_future(future.uuid)
        self._trigger_plotting(future, scan, background)
        self._spawn(
            self._consume_future_result, future, self._scan_number_from_scan(scan)
        )



"""Send a workflow to the ewoks Celery application and recieve the
intermediate results (ewoks events) or final result (job return value)
"""

import os
import tempfile
from typing import List

from ewoksjob.client import submit
from ewoksjob.events.readers import instantiate_reader

from .. import EWOKS_EVENTS_DIR
from .. import testing


def submit_demo():
    test_submit_with_events_in_redis()
    test_submit_with_events_in_sqlite()


@testing.integration_test
def test_submit_with_events_in_redis():
    events_url = "redis://localhost:10002/5"
    handlers = [
        {
            "class": "ewoksjob.events.handlers.RedisEwoksEventHandler",
            "arguments": [{"name": "url", "value": events_url}],
        }
    ]
    _test_submit_with_events(events_url, handlers)


@testing.integration_test
def test_submit_with_events_in_sqlite():
    events_url = f"file://{EWOKS_EVENTS_DIR}/ewoks_event.db"
    handlers = [
        {
            "class": "ewoksjob.events.handlers.Sqlite3EwoksEventHandler",
            "arguments": [{"name": "uri", "value": events_url}],
        }
    ]
    _test_submit_with_events(events_url, handlers)


def _test_submit_with_events(events_url: str, handlers: List[dict]):
    # Test workflow to execute
    workflow = {
        "graph": {"id": "mygraph"},
        "nodes": [
            {"id": "task1", "task_type": "method", "task_identifier": "numpy.add"},
            {"id": "task2", "task_type": "method", "task_identifier": "numpy.add"},
        ],
        "links": [
            {
                "source": "task1",
                "target": "task2",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            }
        ],
    }

    # Job arguments
    out_dir = tempfile.mkdtemp(dir=os.environ["DEMO_TMP_ROOT"], prefix="test_job_")
    varinfo = {"root_uri": out_dir, "scheme": "nexus"}
    inputs = [
        {"id": "task1", "name": 0, "value": 1},
        {"id": "task1", "name": 1, "value": 2},
        {"id": "task2", "name": 1, "value": 3},
    ]
    execinfo = {"handlers": handlers}
    args = (workflow,)
    kwargs = {
        "engine": None,
        "execinfo": execinfo,
        "inputs": inputs,
        "varinfo": varinfo,
        "outputs": [{"all": False}],
    }

    # Trigger workflow
    future = submit(args=args, kwargs=kwargs)
    job_id = future.uuid

    # events could be received in the mean time (see below)
    workflow_results = future.result(timeout=20, interval=0.1)
    assert workflow_results == {"return_value": 6}

    reader = instantiate_reader(events_url)

    # Get intermediate results from ewoks events
    results_during_execution = list(reader.get_events(job_id=job_id))
    assert len(results_during_execution) == 8  # start/stop for job, workflow and node

    # Get start event of node "task1"
    result_event = list(reader.get_events(job_id=job_id, node_id="task1", type="start"))

    assert len(result_event) == 1
    result_event = result_event[0]

    # Verify output URI's
    if varinfo.get("root_uri"):
        results = result_event["output_uris"]
        assert len(results) == 1
        output_uri = results[0]["value"]
        assert out_dir in output_uri
        testing.assert_equal_hdf5_content(output_uri, {"value": 3})

    # Wait for event to contain small data in-place instead of URI's.
    # Or the Variable should move from ewokscore to ewoksutils.
    #
    # # Get start event of node "task1"
    # result_event = list(
    #     reader.get_events_with_variables(
    #         job_id=job_id, node_id="task1", type="start"
    #     )
    # )
    #
    # assert len(result_event) == 1
    # result_event = result_event[0]
    #
    # # Get access to all output variables of "task1"
    # if varinfo.get("root_uri"):
    #     results = result_event["outputs"]
    #     assert results["return_value"].value == 3



from typing import Any
from typing import Dict
from typing import Optional

from ...bliss_globals import setup_globals
from ...bm08.exafs_plotter import Bm08ExafsProcessor
from ...resources import resource_filename


class DemoBm08ExafsProcessor(Bm08ExafsProcessor):
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> None:
        if defaults is None:
            defaults = {}

        defaults.setdefault("queue", "celery")
        defaults.setdefault("workflow", resource_filename("exafs", "exafs.ows"))
        defaults.setdefault("_scan_type", "Kscan")
        counters = defaults.setdefault("_counters", dict())

        if self._HAS_BLISS:
            energy_unit = setup_globals.energy.unit or "eV"
            counters.setdefault(
                "Kscan",
                {
                    "mu_name": "mu",
                    "mon_name": "mu",
                    "min_log": True,
                    "energy_name": "energy",
                    "energy_unit": energy_unit,
                },
            )

        super().__init__(config=config, defaults=defaults)



from blissoda.demo.tests.itest_bm02 import DemoBm02XrpdProcessor  # noqa: F401
from blissoda.demo.tests.itest_bm02 import bm02_demo_1d  # noqa: F401
from blissoda.demo.tests.itest_bm02 import (  # noqa: F401
    bm02_demo_1d_with_cell_subtraction,
)
from blissoda.demo.tests.itest_bm02 import bm02_demo_2d  # noqa: F401
